"""MQTTClient."""

import asyncio
import importlib.metadata
import inspect
import logging
import sys
from json import dumps, loads
from json.decoder import JSONDecodeError
from typing import Any

import attrs
from paho.mqtt.client import Client, MQTTMessage
from paho.mqtt.enums import CallbackAPIVersion
from paho.mqtt.reasoncodes import ReasonCode

from .device import MQTTDevice, MQTTOrigin, TopicCallback

_LOGGER = logging.getLogger(__name__)


@attrs.define()
class MQTTClient:
    """Basic MQTT Client."""

    devs: list[MQTTDevice]
    availability_topic: str = ""
    origin: MQTTOrigin = attrs.field(
        factory=lambda: MQTTOrigin(
            name="mqtt-entity", sw_version=importlib.metadata.version("mqtt-entity")
        )
    )
    client: Client = attrs.field(init=False, repr=False)
    topics_subscribed: set[str] = attrs.field(init=False, repr=False)
    """All topic we subscribed to."""
    migrate_entities: bool = attrs.field(default=True)
    """Migrate entities on discovery info publish. Else remove them."""

    def __attrs_post_init__(self) -> None:
        """Init MQTT Client."""
        self.topics_subscribed = set()
        self.client = Client(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=self.origin.name,
        )
        self.client.on_connect = _mqtt_on_connect
        self.client.on_message = _mqtt_on_message

    async def connect(
        self,
        options: Any = None,
        *,
        username: str | None = None,
        password: str | None = None,
        host: str = "homeassistant.local",
        port: int = 1883,
    ) -> None:
        """Connect to MQTT server specified as attributes of the options."""
        if self.client.is_connected():
            return
        # Disconnect so that we trigger "Connection Successful" on re-connect
        await self.disconnect()

        username = getattr(options, "mqtt_username", username)
        password = getattr(options, "mqtt_password", password)
        host = getattr(options, "mqtt_host", host)
        port = getattr(options, "mqtt_port", port)
        self.client.username_pw_set(username=username, password=password)

        if self.availability_topic:
            self.client.will_set(self.availability_topic, "offline", retain=True)

        _LOGGER.info("MQTT: Connecting to %s@%s:%s", username, host, port)
        self.client.connect_async(host=host, port=port)
        self.client.loop_start()

        retry = 10
        while retry and not self.client.is_connected():
            await asyncio.sleep(0.5)
            retry -= 0
        if not retry:
            raise ConnectionError(
                f"MQTT: Could not connect to {username}@{host}:{port}"
            )
        # publish online (Last will sets offline on disconnect)
        if self.availability_topic:
            await self.publish(self.availability_topic, "online", retain=True)

        # Ensure we subscribe all existing change handlers (after a reconnect)
        for topic in self.topics_subscribed:
            self.client.subscribe(topic)

    async def disconnect(self) -> None:
        """Stop the MQTT client."""

        def _stop() -> None:
            """Do not disconnect, we want the broker to always publish will."""
            self.client.loop_stop()

        await asyncio.get_running_loop().run_in_executor(None, _stop)

    async def publish(
        self,
        topic: str,
        payload: str | None = None,
        qos: int = 0,
        retain: bool = False,
    ) -> None:
        """Publish a MQTT message."""
        if not topic:
            raise ValueError(f"MQTT: Cannot publish to empty topic (payload={payload})")
        if not isinstance(qos, int):
            qos = 0
        if retain:
            qos = 1
        _LOGGER.debug(
            "MQTT: Publish %s%s %s, %s", qos, "R" if retain else "", topic, payload
        )
        await asyncio.get_running_loop().run_in_executor(
            None, self.client.publish, topic, payload, qos, bool(retain)
        )

    def publish_discovery_info(self) -> None:
        """Publish discovery info if HA is online."""

        async def _timeout() -> None:
            """Timeout for Home Assistant online check."""
            _LOGGER.error(
                "MQTT: Home Assistant not online. Topic homeassistant/status is empty"
            )
            sys.exit(1)

        timeout = asyncio.get_running_loop().call_later(10, _timeout)

        async def _online_cb(payload_s: str) -> None:
            """Republish discovery info."""
            if payload_s != "online":
                _LOGGER.warning(
                    "MQTT: Home Assistant offline. homeassistant/status = %s",
                    payload_s,
                )
                return

            timeout.cancel()
            timeout.cancel()
            _LOGGER.info("MQTT: Home Assistant online. Publish discovery info.")
            await self._publish_discovery_info()

        if not self.client.is_connected():
            raise ConnectionError()

        self.topic_subscribe("homeassistant/status", _online_cb)

    async def _publish_discovery_info(self) -> None:
        """Home Assistant MQTT discovery helper.

        https://www.home-assistant.io/docs/mqtt/discovery/
        Publish discovery topics on "homeassistant/device/{device_id}/{sensor_id}/config"
        Publish discovery topics on "homeassistant/(sensor|switch)/{device_id}/{sensor_id}/config"
        """

        for ddev in self.devs:
            disco_topic, dico_dict = ddev.discovery_info(
                self.availability_topic, origin=self.origin
            )
            await self.publish(disco_topic, payload=dumps(dico_dict))

            # add topic callbacks
            tcb: dict[str, TopicCallback] = {}
            for ent in ddev.components.values():
                ent.topic_callbacks(tcb)
            for topic, cbk in tcb.items():
                self.topic_subscribe(topic, cbk)

        await asyncio.sleep(0.01)

        if self.migrate_entities:
            _LOGGER.debug("MQTT: Migrate entities")
            await self.clean_entity_based_cb(migrate=True)
        else:
            await self.clean_entity_based_cb(remove=True)

    async def clean_entity_based_cb(
        self, *, migrate: bool = False, remove: bool = False
    ) -> None:
        """Remove previously discovered entities."""

        async def cb_migrate(payload_s: str, topic: str) -> None:
            """Callback to remove the device."""
            if not payload_s:
                return
            payload = mqtt_loads(payload_s)
            _LOGGER.info("MQTT MIGRATE topic %s with payload %s", topic, payload)
            migrate_ok = payload == {"migrate": True}
            _pl = None if migrate_ok else dumps({"migrate": True})
            if migrate_ok:
                await asyncio.sleep(1)
            await self.publish(topic=topic, payload=_pl, qos=1, retain=True)

        def cb_remove(dev: MQTTDevice) -> TopicCallback:
            """Create a callback for the device."""

            async def _cb_remove(payload_s: str, topic: str) -> None:
                if not payload_s:
                    return
                payload = mqtt_loads(payload_s)
                # if not part of this device, remove the topic
                if not isinstance(payload, dict) or "unique_id" not in payload:
                    _LOGGER.warning(
                        "MQTT CLEAN: No unique_id in payload %s, cannot remove", payload
                    )
                    return
                uid = payload["unique_id"]
                if uid not in dev.components:
                    _LOGGER.info("MQTT: Removing unique ID %s", uid)
                    self.client.publish(topic=topic, payload=None, qos=1, retain=True)

            return _cb_remove

        for dev in self.devs:
            topic = f"homeassistant/+/{dev.id}/+/config"
            self.topic_subscribe(topic, cb_migrate if migrate else cb_remove(dev))
            asyncio.get_running_loop().call_later(10, self.topic_unsubscribe, topic)

    def topic_unsubscribe(self, topic: str) -> None:
        """Remove a topic from the topic callbacks."""
        self.client.unsubscribe(topic)
        self.client.message_callback_remove(topic)
        self.topics_subscribed.discard(topic)

    def topic_subscribe(self, topic: str, callback: TopicCallback) -> None:
        """Add a topic to the topic callbacks."""

        paramc = len(inspect.signature(callback).parameters)
        loop = asyncio.get_running_loop()

        def cb(_client: Client, _userdata: Any, message: MQTTMessage) -> None:
            """Callback for the topic."""
            payload = message.payload.decode("utf-8")
            _LOGGER.debug(
                "MQTT Callback for topic %s, payload %s",
                message.topic,
                payload or '""',
            )
            args = (payload,) if paramc == 1 else (payload, message.topic)
            if inspect.iscoroutinefunction(callback):
                coro = callback(*args)
                loop.call_soon_threadsafe(lambda: loop.create_task(coro))
            else:
                callback(*args)

        _LOGGER.debug("MQTT add callback for topic %s", topic)
        self.client.subscribe(topic)
        self.client.message_callback_add(topic, cb)
        self.topics_subscribed.add(topic)


def mqtt_loads(msg: str) -> dict[str, Any] | str:
    """Load a JSON string into a dictionary."""
    if msg is None:
        return {}
    try:
        res = loads(msg)
        if isinstance(res, dict):
            return res
    except JSONDecodeError:
        pass
    return str(msg)


def _mqtt_on_connect(
    _client: Client,
    _userdata: Any,
    _flags: Any,
    _rc: ReasonCode,
    _prop: Any = None,
) -> None:
    """MQTT on_connect callback."""
    if _rc == 0:
        _LOGGER.info("MQTT: Connection successful")
        return
    _LOGGER.error("MQTT: Connection failed with reason code %s", _rc)

    # msg = {
    #     mqttc.CONNACK_ACCEPTED: "successful",
    #     mqttc.CONNACK_REFUSED_PROTOCOL_VERSION: "refused - incorrect protocol version",
    #     mqttc.CONNACK_REFUSED_IDENTIFIER_REJECTED: "refused - invalid client identifier",
    #     mqttc.CONNACK_REFUSED_SERVER_UNAVAILABLE: "refused - server unavailable",
    #     mqttc.CONNACK_REFUSED_BAD_USERNAME_PASSWORD: "refused - bad username or password",
    #     mqttc.CONNACK_REFUSED_NOT_AUTHORIZED: "refused - not authorised",
    # }.get(_rc, f"refused - {_rc}")
    # _LOGGER.info("MQTT: Connection %s", msg)


def _mqtt_on_message(
    _client: Client,
    _userdata: Any,
    message: MQTTMessage,
) -> None:
    """MQTT on_message callback."""
    topic = message.topic
    payload = message.payload.decode("utf-8")
    _LOGGER.warning(
        "MQTT: Unhandled msg received on topic %s with payload %s", topic, payload
    )



"""HASS MQTT Device, used for device based discovery."""

import inspect
import logging
from typing import Any, Callable, Coroutine

import attrs
from attrs import validators

_LOGGER = logging.getLogger(__name__)


type TopicCallback = (
    Callable[[str], None]
    | Callable[[str, str], None]
    | Callable[[str], Coroutine[Any, Any, None]]
    | Callable[[str, str], Coroutine[Any, Any, None]]
)


@attrs.define()
class MQTTBaseEntity:
    """Base class for entities that support MQTT Discovery."""

    def discovery_dict(self, result: dict[str, Any]) -> None:
        """Post-process the discovery dictionary."""

    def topic_callbacks(self, result: dict[str, TopicCallback]) -> None:
        """Append topics and callbacks."""


@attrs.define()
class MQTTOrigin:
    """Represent the origin of an MQTT message."""

    name: str
    sw_version: str = ""
    support_url: str = ""


@attrs.define()
class MQTTDevice:
    """Base class for MQTT Device Discovery. A Home Assistant Device groups entities."""

    components: dict[str, MQTTBaseEntity]
    """MQTT component entities."""

    identifiers: list[str | tuple[str, Any]] = attrs.field(
        validator=[validators.instance_of(list), validators.min_len(1)]
    )
    connections: list[str] = attrs.field(factory=list)
    configuration_url: str = ""
    manufacturer: str = ""
    model: str = ""
    name: str = ""
    suggested_area: str = ""
    sw_version: str = ""
    via_device: str = ""

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """The device identifier."""
        return str(self.identifiers[0])

    def discovery_info(
        self, availability_topic: str, *, origin: MQTTOrigin
    ) -> tuple[str, dict[str, Any]]:
        """Return the discovery dictionary for the MQTT device."""
        return (
            f"homeassistant/device/{self.id}/config",
            {
                "dev": discovery_dict(self, exclude=["components"]),
                "o": discovery_dict(origin),
                "avty": {
                    "topic": availability_topic,
                },
                "cmps": {k: discovery_dict(v) for k, v in self.components.items()},
            },
        )


def discovery_dict(
    obj: attrs.AttrsInstance, exclude: list[str] | None = None
) -> dict[str, Any]:
    """Represent the entity as a dictionary, without empty values and defaults."""

    def _filter(atrb: attrs.Attribute, value: Any) -> bool:
        if exclude and atrb.name in exclude:
            return False
        if atrb.name in ("discovery_extra", "_path"):
            return False
        return bool(value) and atrb.default != value and not inspect.isfunction(value)

    res = attrs.asdict(obj, filter=_filter)

    extra = getattr(obj, "discovery_extra", None)
    if extra:
        keys = {
            key: extra[key] for key in extra if key in res and res[key] != extra[key]
        }
        _LOGGER.debug("Overwriting %s", keys)
        res.update(extra)

    if isinstance(obj, MQTTBaseEntity):
        obj.discovery_dict(res)
        if not res.get("platform"):  # Required for device based discovery
            res["platform"] = getattr(obj, "_path", "")

    return res



"""mqtt_entity library."""

from mqtt_entity.client import MQTTClient  # noqa
from mqtt_entity.device import MQTTDevice, MQTTOrigin, MQTTBaseEntity  # noqa
from mqtt_entity.entities import (  # noqa
    MQTTBinarySensorEntity,
    MQTTDeviceTrigger,
    MQTTEntity,
    MQTTLightEntity,
    MQTTNumberEntity,
    MQTTRWEntity,
    MQTTSelectEntity,
    MQTTSensorEntity,
    MQTTSwitchEntity,
)
from mqtt_entity.helpers import (  # noqa
    hass_default_rw_icon,
    hass_device_class,
)



"""MQTT entities."""

from __future__ import annotations

from json import dumps
from typing import TYPE_CHECKING, Any, Callable, Sequence

import attrs
from attrs import validators

from .device import MQTTBaseEntity, TopicCallback
from .utils import BOOL_OFF, BOOL_ON, required

if TYPE_CHECKING:
    from .client import MQTTClient

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@attrs.define()
class MQTTEntity(MQTTBaseEntity):
    """A generic Home Assistant entity used as the base class for other entities."""

    unique_id: str
    # device: MQTTDevice
    state_topic: str
    name: str
    object_id: str = ""
    # availability: list[MQTTAvailability] = attrs.field(factory=list)
    # availability_mode: str = ""
    device_class: str = ""
    unit_of_measurement: str = ""
    state_class: str = ""
    expire_after: int = 0
    """Unavailable if not updated."""
    enabled_by_default: bool = True
    entity_category: str = ""
    entity_picture: str = ""
    icon: str = ""
    json_attributes_topic: str = ""
    """Used by the set_attributes helper."""

    discovery_extra: dict[str, Any] = attrs.field(factory=dict)
    """Additional MQTT Discovery attributes."""

    _path = ""

    async def send_state(
        self, client: MQTTClient, payload: str, *, retain: bool = False
    ) -> None:
        """Publish the state to the MQTT state topic."""
        await client.publish(self.state_topic, payload, retain=retain)

    async def send_json_attributes(
        self, client: MQTTClient, attributes: dict[str, Any], *, retain: bool = True
    ) -> None:
        """Publish the attributes to the MQTT attributes topic."""
        await client.publish(
            topic=self.json_attributes_topic, payload=dumps(attributes), retain=retain
        )

    def __attrs_post_init__(self) -> None:
        """Init the class."""
        if not self._path:
            raise TypeError(f"Do not instantiate {self.__class__.__name__} directly")
        if not self.state_class and self.device_class == "energy":
            self.state_class = "total_increasing"


@attrs.define()
class MQTTDeviceTrigger(MQTTBaseEntity):
    """A Home Assistant Device trigger.

    https://www.home-assistant.io/integrations/device_trigger.mqtt/
    """

    type: str
    subtype: str
    payload: str
    topic: str

    _path = "device_automation"

    @property
    def name(self) -> str:
        """Return the name of the trigger."""
        return f"{self.type} {self.subtype}".strip()

    async def send_trigger(self, client: MQTTClient) -> None:
        """Publish the state to the MQTT state topic."""
        await client.publish(self.topic, self.payload or "1")

    discovery_extra: dict[str, Any] = attrs.field(factory=dict)
    """Additional MQTT Discovery attributes."""

    def discovery_dict(self, result: dict[str, Any]) -> None:
        """Return the final discovery dictionary."""
        super().discovery_dict(result)
        result["automation_type"] = "trigger"
        result["platform"] = "device_automation"

    # on_trigger: Callable | None = None
    # """Callable to call when triggered."""

    # def topic_callbacks(self, result: dict[str, TopicCallback]) -> None:
    #     """Return a dictionary of topic callbacks."""
    #     super().topic_callbacks(result)
    #     if self.topic and self.on_trigger:
    #         result[self.topic] = self.on_trigger


@attrs.define()
class MQTTRWEntity(MQTTEntity):
    """Read/Write entity base class.

    This will default to a text entity.
    """

    command_topic: str = attrs.field(
        default="", validator=(validators.instance_of(str), validators.min_len(2))
    )
    on_command: Callable | None = None

    _path = "text"

    def topic_callbacks(self, result: dict[str, TopicCallback]) -> None:
        """Return a dictionary of topic callbacks."""
        super().topic_callbacks(result)
        if self.command_topic and self.on_command:
            result[self.command_topic] = self.on_command


@attrs.define()
class MQTTSensorEntity(MQTTEntity):
    """A Home Assistant Sensor entity."""

    _path = "sensor"


@attrs.define()
class MQTTBinarySensorEntity(MQTTEntity):
    """A Home Assistant Binary Sensor entity."""

    payload_on: str = BOOL_ON
    payload_off: str = BOOL_OFF

    _path = "binary_sensor"


@attrs.define()
class MQTTSelectEntity(MQTTRWEntity):
    """A HomeAssistant Select entity."""

    options: Sequence[str] = attrs.field(default=None, validator=required)

    _path = "select"


@attrs.define()
class MQTTSwitchEntity(MQTTRWEntity):
    """A Home Assistant Switch entity."""

    payload_on: str = BOOL_ON
    payload_off: str = BOOL_OFF

    _path = "switch"


@attrs.define()
class MQTTText(MQTTRWEntity):
    """A Home Assistant Switch entity."""

    _path = "text"


@attrs.define()
class MQTTLightEntity(MQTTRWEntity):
    """A Home Assistant Switch entity."""

    payload_on: str = BOOL_ON
    payload_off: str = BOOL_OFF

    brightness_state_topic: str = ""
    brightness_command_topic: str = ""
    on_brightness_change: TopicCallback | None = None

    effect_state_topic: str = ""
    effect_command_topic: str = ""
    on_effect_change: TopicCallback | None = None
    effect_list: list[str] | None = None

    hs_state_topic: str = ""
    hs_command_topic: str = ""
    on_hs_change: TopicCallback | None = None

    _path = "light"

    async def send_brightness(
        self, client: MQTTClient, brightness: int, *, retain: bool = False
    ) -> None:
        """Publish the brightness to the MQTT brightness command topic."""
        await client.publish(
            self.brightness_state_topic,
            str(brightness),
            retain=retain,
        )

    async def send_effect(
        self, client: MQTTClient, effect: str, *, retain: bool = False
    ) -> None:
        """Publish the effect to the MQTT effect command topic."""
        await client.publish(
            self.effect_state_topic,
            effect,
            retain=retain,
        )

    async def send_hs(
        self, client: MQTTClient, hs: tuple[float, float], *, retain: bool = False
    ) -> None:
        """Publish the hue and saturation to the MQTT hs command topic."""
        await client.publish(
            self.hs_state_topic,
            f"{hs[0]},{hs[1]}",
            retain=retain,
        )

    def topic_callbacks(self, result: dict[str, TopicCallback]) -> None:
        """Return a dictionary of topic callbacks."""
        super().topic_callbacks(result)
        if self.brightness_command_topic and self.on_brightness_change:
            result[self.brightness_command_topic] = self.on_brightness_change
        if self.effect_command_topic and self.on_effect_change:
            result[self.effect_command_topic] = self.on_effect_change
        if self.hs_command_topic and self.on_hs_change:
            result[self.hs_command_topic] = self.on_hs_change


@attrs.define()
class MQTTNumberEntity(MQTTRWEntity):
    """A HomeAssistant Number entity."""

    min: float = 0.0
    max: float = 100.0
    mode: str = "auto"
    step: float = 1.0

    _path = "number"
